@@ -510,15 +510,23 @@ get_access_technology (MMGenericGsm *modem,
 /*****************************************************************************/
 
 static gboolean
-parse_num (const char *str, guint32 *out_num, guint32 min, guint32 max)
+parse_match_as_num (GMatchInfo *match_info,
+                    guint idx,
+                    guint32 *out_num,
+                    guint32 min,
+                    guint32 max)
 {
     unsigned long int tmp;
+    char *str;
 
+    str = g_match_info_fetch (match_info, idx);
     if (!str || !strlen (str))
         return FALSE;
 
     errno = 0;
     tmp = strtoul (str, NULL, 10);
+    g_free (str);
+
     if (errno != 0 || tmp < min || tmp > max)
         return FALSE;
     *out_num = (guint32) tmp;
@@ -534,10 +542,12 @@ send_huawei_cpin_done (MMAtSerialPort *port,
     MMCallbackInfo *info = (MMCallbackInfo *) user_data;
     GRegex *r = NULL;
     GMatchInfo *match_info = NULL;
-    const char *pin_type;
-    guint32 attempts_left = 0;
-    char *str = NULL;
-    guint32 num = 0;
+    guint32 i = 0;
+    gboolean success = FALSE;
+    GArray *retry_counts;
+    PinRetryCount ur[4] = {
+        {"sim-puk", 0}, {"sim-pin", 0}, {"sim-puk2", 0}, {"sim-pin2", 0}
+    };
 
     /* If the modem has already been removed, return without
      * scheduling callback */
@@ -549,9 +559,7 @@ send_huawei_cpin_done (MMAtSerialPort *port,
         goto done;
     }
 
-    pin_type = mm_callback_info_get_data (info, "pin_type");
-
-	r = g_regex_new ("\\^CPIN:\\s*([^,]+),[^,]*,(\\d+),(\\d+),(\\d+),(\\d+)", G_REGEX_UNGREEDY, 0, NULL);
+    r = g_regex_new ("\\^CPIN:\\s*([^,]+),[^,]*,(\\d+),(\\d+),(\\d+),(\\d+)", G_REGEX_UNGREEDY, 0, NULL);
     if (!r) {
         g_set_error_literal (&info->error,
                              MM_MODEM_ERROR, MM_MODEM_ERROR_GENERAL,
@@ -565,37 +573,21 @@ send_huawei_cpin_done (MMAtSerialPort *port,
         goto done;
     }
 
-    if (strstr (pin_type, MM_MODEM_GSM_CARD_SIM_PUK))
-        num = 2;
-    else if (strstr (pin_type, MM_MODEM_GSM_CARD_SIM_PIN))
-        num = 3;
-    else if (strstr (pin_type, MM_MODEM_GSM_CARD_SIM_PUK2))
-        num = 4;
-    else if (strstr (pin_type, MM_MODEM_GSM_CARD_SIM_PIN2))
-        num = 5;
-    else {
-        mm_dbg ("unhandled pin type '%s'", pin_type);
-
-        info->error = g_error_new_literal (MM_MODEM_ERROR, MM_MODEM_ERROR_GENERAL, "Unhandled PIN type");
-    }
-
-    if (num > 0) {
-        gboolean success = FALSE;
-
-        str = g_match_info_fetch (match_info, num);
-        if (str) {
-            success = parse_num (str, &attempts_left, 0, 10);
-            g_free (str);
-        }
-
+    for (i = 0; i <= 3; i++) {
+        success = parse_match_as_num (match_info, i + 2, &ur[i].count, 0, 10);
         if (!success) {
             info->error = g_error_new_literal (MM_MODEM_ERROR,
                                                MM_MODEM_ERROR_GENERAL,
                                                "Could not parse ^CPIN results (missing or invalid match info).");
+            break;
         }
     }
 
-    mm_callback_info_set_result (info, GUINT_TO_POINTER (attempts_left), NULL);
+    if (success) {
+        retry_counts = g_array_sized_new (FALSE, TRUE, sizeof (PinRetryCount), 4);
+        g_array_append_vals (retry_counts, &ur, 4);
+        mm_callback_info_set_result (info, retry_counts, NULL);
+    }
 
 done:
     if (match_info)
@@ -608,15 +600,12 @@ done:
 
 static void
 get_unlock_retries (MMModemGsmCard *modem,
-                    const char *pin_type,
-                    MMModemUIntFn callback,
+                    MMModemArrayFn callback,
                     gpointer user_data)
 {
     MMAtSerialPort *port;
     char *command;
-    MMCallbackInfo *info = mm_callback_info_uint_new (MM_MODEM (modem), callback, user_data);
-
-    mm_dbg ("pin type '%s'", pin_type);
+    MMCallbackInfo *info = mm_callback_info_array_new (MM_MODEM (modem), callback, user_data);
 
     /* Ensure we have a usable port to use for the command */
     port = mm_generic_gsm_get_best_at_port (MM_GENERIC_GSM (modem), &info->error);
@@ -639,8 +628,6 @@ get_unlock_retries (MMModemGsmCard *modem,
     mm_at_serial_port_queue_command (port, command, 3, NULL, NULL);
     g_free (command);
 
-    mm_callback_info_set_data (info, "pin_type", g_strdup (pin_type), g_free);
-
     command = g_strdup_printf ("^CPIN?");
     mm_at_serial_port_queue_command (port, command, 3, send_huawei_cpin_done, info);
     g_free (command);
@@ -727,81 +714,114 @@ handle_status_change (MMAtSerialPort *port,
 
 /*****************************************************************************/
 
-static gboolean
-grab_port (MMModem *modem,
-           const char *subsys,
-           const char *name,
-           MMPortType suggested_type,
-           gpointer user_data,
-           GError **error)
+static void
+do_enable_power_up_done (MMGenericGsm *gsm,
+                         GString *response,
+                         GError *error,
+                         MMCallbackInfo *info)
 {
-    MMGenericGsm *gsm = MM_GENERIC_GSM (modem);
-    MMPortType ptype = MM_PORT_TYPE_IGNORED;
-    const char *sys[] = { "tty", NULL };
-    GUdevClient *client;
-    GUdevDevice *device = NULL;
-    MMPort *port = NULL;
-    int usbif;
+    if (!error) {
+        MMAtSerialPort *primary;
 
-    client = g_udev_client_new (sys);
-    if (!client) {
-        g_set_error (error, 0, 0, "Could not get udev client.");
-        return FALSE;
+        /* Enable unsolicited result codes */
+        primary = mm_generic_gsm_get_at_port (gsm, MM_AT_PORT_FLAG_PRIMARY);
+        g_assert (primary);
+
+        mm_at_serial_port_queue_command (primary, "^PORTSEL=0", 5, NULL, NULL);
+        mm_at_serial_port_queue_command (primary, "^CURC=1", 5, NULL, NULL);
     }
 
-    device = g_udev_client_query_by_subsystem_and_name (client, subsys, name);
-    if (!device) {
-        g_set_error (error, 0, 0, "Could not get udev device.");
-        goto out;
+    /* Chain up to parent */
+    MM_GENERIC_GSM_CLASS (mm_modem_huawei_gsm_parent_class)->do_enable_power_up_done (gsm, NULL, error, info);
+}
+
+/*****************************************************************************/
+
+static void
+disable_unsolicited_done (MMAtSerialPort *port,
+                          GString *response,
+                          GError *error,
+                          gpointer user_data)
+
+{
+    MMCallbackInfo *info = (MMCallbackInfo *) user_data;
+
+    /* If the modem has already been removed, return without
+     * scheduling callback */
+    if (mm_callback_info_check_modem_removed (info))
+        return;
+
+    /* Ignore all errors */
+    mm_callback_info_schedule (info);
+}
+
+static void
+invoke_call_parent_disable_fn (MMCallbackInfo *info)
+{
+    /* Note: we won't call the parent disable if info->modem is no longer
+     * valid. The invoke is called always once the info gets scheduled, which
+     * may happen during removed modem detection. */
+    if (info->modem) {
+        MMModem *parent_modem_iface;
+
+        parent_modem_iface = g_type_interface_peek_parent (MM_MODEM_GET_INTERFACE (info->modem));
+        parent_modem_iface->disable (info->modem, (MMModemFn)info->callback, info->user_data);
     }
+}
 
-    usbif = g_udev_device_get_property_as_int (device, "ID_USB_INTERFACE_NUM");
-    if (usbif < 0) {
-        g_set_error (error, 0, 0, "Could not get USB device interface number.");
-        goto out;
-    }
+static void
+disable (MMModem *modem,
+         MMModemFn callback,
+         gpointer user_data)
+{
+    MMAtSerialPort *primary;
+    MMCallbackInfo *info;
 
-    if (usbif == 0) {
-        if (!mm_generic_gsm_get_at_port (gsm, MM_PORT_TYPE_PRIMARY))
-            ptype = MM_PORT_TYPE_PRIMARY;
-    } else if (suggested_type == MM_PORT_TYPE_SECONDARY) {
-        if (!mm_generic_gsm_get_at_port (gsm, MM_PORT_TYPE_SECONDARY))
-            ptype = MM_PORT_TYPE_SECONDARY;
-    }
+    info = mm_callback_info_new_full (modem,
+                                      invoke_call_parent_disable_fn,
+                                      (GCallback)callback,
+                                      user_data);
 
-    port = mm_generic_gsm_grab_port (gsm, subsys, name, ptype, error);
+    primary = mm_generic_gsm_get_at_port (MM_GENERIC_GSM (modem), MM_AT_PORT_FLAG_PRIMARY);
+    g_assert (primary);
 
-    if (port && MM_IS_AT_SERIAL_PORT (port)) {
-        GRegex *regex;
+    /* Turn off unsolicited responses */
+    mm_at_serial_port_queue_command (primary, "^CURC=0", 5, disable_unsolicited_done, info);
+}
 
+/*****************************************************************************/
+
+static void
+port_grabbed (MMGenericGsm *gsm,
+              MMPort *port,
+              MMAtPortFlags pflags,
+              gpointer user_data)
+{
+    GRegex *regex;
+
+    if (MM_IS_AT_SERIAL_PORT (port)) {
         g_object_set (G_OBJECT (port), MM_PORT_CARRIER_DETECT, FALSE, NULL);
 
         regex = g_regex_new ("\\r\\n\\^RSSI:(\\d+)\\r\\n", G_REGEX_RAW | G_REGEX_OPTIMIZE, 0, NULL);
-        mm_at_serial_port_add_unsolicited_msg_handler (MM_AT_SERIAL_PORT (port), regex, handle_signal_quality_change, modem, NULL);
+        mm_at_serial_port_add_unsolicited_msg_handler (MM_AT_SERIAL_PORT (port), regex, handle_signal_quality_change, gsm, NULL);
         g_regex_unref (regex);
 
         regex = g_regex_new ("\\r\\n\\^MODE:(\\d),(\\d)\\r\\n", G_REGEX_RAW | G_REGEX_OPTIMIZE, 0, NULL);
-        mm_at_serial_port_add_unsolicited_msg_handler (MM_AT_SERIAL_PORT (port), regex, handle_mode_change, modem, NULL);
+        mm_at_serial_port_add_unsolicited_msg_handler (MM_AT_SERIAL_PORT (port), regex, handle_mode_change, gsm, NULL);
         g_regex_unref (regex);
 
         regex = g_regex_new ("\\r\\n\\^DSFLOWRPT:(.+)\\r\\n", G_REGEX_RAW | G_REGEX_OPTIMIZE, 0, NULL);
-        mm_at_serial_port_add_unsolicited_msg_handler (MM_AT_SERIAL_PORT (port), regex, handle_status_change, modem, NULL);
+        mm_at_serial_port_add_unsolicited_msg_handler (MM_AT_SERIAL_PORT (port), regex, handle_status_change, gsm, NULL);
         g_regex_unref (regex);
 
         regex = g_regex_new ("\\r\\n\\^BOOT:.+\\r\\n", G_REGEX_RAW | G_REGEX_OPTIMIZE, 0, NULL);
-        mm_at_serial_port_add_unsolicited_msg_handler (MM_AT_SERIAL_PORT (port), regex, NULL, modem, NULL);
+        mm_at_serial_port_add_unsolicited_msg_handler (MM_AT_SERIAL_PORT (port), regex, NULL, gsm, NULL);
         g_regex_unref (regex);
 
         regex = g_regex_new ("\\r\\n\\^CSNR:.+\\r\\n", G_REGEX_RAW | G_REGEX_OPTIMIZE, 0, NULL);
-        mm_at_serial_port_add_unsolicited_msg_handler (MM_AT_SERIAL_PORT (port), regex, NULL, modem, NULL);
+        mm_at_serial_port_add_unsolicited_msg_handler (MM_AT_SERIAL_PORT (port), regex, NULL, gsm, NULL);
         g_regex_unref (regex);
     }
-
-out:
-    if (device)
-        g_object_unref (device);
-    g_object_unref (client);
-    return !!port;
 }
 
 /* Encode to packed GSM - this is what Huawei supports on all known models */
@@ -858,7 +878,7 @@ ussd_decode (MMModemGsmUssd *self, const char* reply, guint scheme)
 static void
 modem_init (MMModem *modem_class)
 {
-    modem_class->grab_port = grab_port;
+    modem_class->disable = disable;
 }
 
 static void
@@ -895,7 +915,9 @@ mm_modem_huawei_gsm_class_init (MMModemHuaweiGsmClass *klass)
     mm_modem_huawei_gsm_parent_class = g_type_class_peek_parent (klass);
     g_type_class_add_private (object_class, sizeof (MMModemHuaweiGsmPrivate));
 
+    gsm_class->port_grabbed = port_grabbed;
     gsm_class->set_allowed_mode = set_allowed_mode;
     gsm_class->get_allowed_mode = get_allowed_mode;
     gsm_class->get_access_technology = get_access_technology;
+    gsm_class->do_enable_power_up_done = do_enable_power_up_done;
 }
