@@ -31,7 +31,6 @@
 #include "mm-modem-gsm-card.h"
 #include "mm-log.h"
 #include "mm-modem-icera.h"
-#include "mm-utils.h"
 #include "mm-modem-time.h"
 
 static void modem_init (MMModem *modem_class);
@@ -53,9 +52,8 @@ G_DEFINE_TYPE_EXTENDED (MMModemSamsungGsm, mm_modem_samsung_gsm, MM_TYPE_GENERIC
 
 typedef struct {
     gboolean disposed;
-
     MMModemIceraPrivate *icera;
-    char *band;
+    MMModemGsmBand supported_bands;
 } MMModemSamsungGsmPrivate;
 
 #define IPDPADDR_TAG "%IPDPADDR: "
@@ -84,44 +82,6 @@ mm_modem_samsung_gsm_new (const char *device,
     return modem;
 }
 
-typedef struct {
-    MMModemGsmBand mm;
-    char band[50];
-} BandTable;
-
-static BandTable bands[12] = {
-    /* Sort 3G first since it's preferred */
-    { MM_MODEM_GSM_BAND_U2100, "FDD_BAND_I" },
-    { MM_MODEM_GSM_BAND_U1900, "FDD_BAND_II" },
-    { MM_MODEM_GSM_BAND_U1800, "FDD_BAND_III" },
-    { MM_MODEM_GSM_BAND_U17IV, "FDD_BAND_IV" },
-    { MM_MODEM_GSM_BAND_U850,  "FDD_BAND_V" },
-    { MM_MODEM_GSM_BAND_U800,  "FDD_BAND_VI" },
-    { MM_MODEM_GSM_BAND_U900,  "FDD_BAND_VIII" },
-    { MM_MODEM_GSM_BAND_G850, "G850" },
-    /* 2G second */
-    { MM_MODEM_GSM_BAND_DCS,   "DCS" },
-    { MM_MODEM_GSM_BAND_EGSM,  "EGSM" }, /* 0x100 = Extended GSM, 0x200 = Primary GSM */
-    { MM_MODEM_GSM_BAND_PCS,   "PCS" },
-    /* And ANY last since it's most inclusive */
-    { MM_MODEM_GSM_BAND_ANY,   "ANY" },
-};
-
-static gboolean
-band_mm_to_samsung (MMModemGsmBand band, MMModemGsmNetwork *modem)
-{
-    int i;
-    MMModemSamsungGsmPrivate *priv = MM_MODEM_SAMSUNG_GSM_GET_PRIVATE (modem);
-
-    for (i = 0; i < sizeof (bands) / sizeof (BandTable); i++) {
-        if (bands[i].mm == band) {
-            priv->band = bands[i].band;
-            return TRUE;
-        }
-    }
-    return FALSE;
-}
-
 static void
 get_allowed_mode (MMGenericGsm *gsm,
                   MMModemUIntFn callback,
@@ -139,118 +99,13 @@ set_allowed_mode (MMGenericGsm *gsm,
     mm_modem_icera_set_allowed_mode (MM_MODEM_ICERA (gsm), mode, callback, user_data);
 }
 
-static void
-set_band_done (MMAtSerialPort *port,
-               GString *response,
-               GError *error,
-               gpointer user_data)
-{
-    MMCallbackInfo *info = (MMCallbackInfo *) user_data;
-
-    /* If the modem has already been removed, return without
-     * scheduling callback */
-    if (mm_callback_info_check_modem_removed (info))
-        return;
-
-    if (error)
-        info->error = g_error_copy (error);
-
-    mm_callback_info_schedule (info);
-}
-
 static void
 set_band (MMModemGsmNetwork *modem,
           MMModemGsmBand band,
           MMModemFn callback,
           gpointer user_data)
 {
-    MMCallbackInfo *info;
-    MMAtSerialPort *port;
-    char *command;
-    MMModemSamsungGsmPrivate *priv = MM_MODEM_SAMSUNG_GSM_GET_PRIVATE (modem);
-
-    info = mm_callback_info_new (MM_MODEM (modem), callback, user_data);
-
-    port = mm_generic_gsm_get_best_at_port (MM_GENERIC_GSM (modem), &info->error);
-    if (!port) {
-        mm_callback_info_schedule (info);
-        return;
-    }
-
-    /* TODO: Check how to pass more than one band in the same AT%%IPBM command */
-    if (!utils_check_for_single_value (band)) {
-        info->error = g_error_new_literal (MM_MODEM_ERROR, MM_MODEM_ERROR_GENERAL, "Cannot set more than one band.");
-        mm_callback_info_schedule (info);
-    } else if (!band_mm_to_samsung (band, modem)) {
-        info->error = g_error_new_literal (MM_MODEM_ERROR, MM_MODEM_ERROR_GENERAL, "Invalid band.");
-        mm_callback_info_schedule (info);
-    } else {
-        mm_callback_info_set_data (info, "band", g_strdup(priv->band), NULL);
-        command = g_strdup_printf ("AT%%IPBM=\"%s\",1", priv->band);
-        mm_at_serial_port_queue_command (port, command, 3, set_band_done, info);
-        g_free (command);
-        priv->band = NULL;
-    }
-}
-
-static gboolean
-parse_ipbm (const char *reply, MMModemGsmBand *band)
-{
-    int enable[12];
-
-    g_return_val_if_fail (band != NULL, FALSE);
-    g_return_val_if_fail (reply != NULL, FALSE);
-
-    if (sscanf (reply, "\"ANY\": %d\r\n\"EGSM\": %d\r\n\"DCS\": %d\r\n\"PCS\": %d\r\n\"G850\": %d\r\n\"FDD_BAND_I\": %d\r\n\"FDD_BAND_II\": %d\r\n\"FDD_BAND_III\": %d\r\n\"FDD_BAND_IV\": %d\r\n\"FDD_BAND_V\": %d\r\n\"FDD_BAND_VI\": %d\r\n\"FDD_BAND_VIII\": %d", &enable[0], &enable[1], &enable[2], &enable[3], &enable[4], &enable[5], &enable[6], &enable[7], &enable[8], &enable[9], &enable[10], &enable[11]) != 12)
-        return FALSE;
-
-    *band = 0;
-    if (enable[5] == 1)
-        *band |= MM_MODEM_GSM_BAND_U2100;
-    if (enable[6] == 1)
-        *band |= MM_MODEM_GSM_BAND_U1900;
-    if (enable[7] == 1)
-        *band |= MM_MODEM_GSM_BAND_U1800;
-    if (enable[8] == 1)
-        *band |= MM_MODEM_GSM_BAND_U17IV;
-    if (enable[9] == 1)
-        *band |= MM_MODEM_GSM_BAND_U850;
-    if (enable[10] == 1)
-        *band |= MM_MODEM_GSM_BAND_U800;
-    if (enable[11] == 1)
-        *band |= MM_MODEM_GSM_BAND_U900;
-    if (enable[1] == 1)
-        *band |= MM_MODEM_GSM_BAND_EGSM;
-    if (enable[2] == 1)
-        *band |= MM_MODEM_GSM_BAND_DCS;
-    if (enable[3] == 1)
-        *band |= MM_MODEM_GSM_BAND_PCS;
-    if (enable[4] == 1)
-        *band |= MM_MODEM_GSM_BAND_G850;
-
-    return (*band > 0 ? TRUE : FALSE);
-}
-
-static void
-get_band_done (MMAtSerialPort *port,
-               GString *response,
-               GError *error,
-               gpointer user_data)
-{
-    MMCallbackInfo *info = (MMCallbackInfo *) user_data;
-    MMModemGsmBand mm_band = MM_MODEM_GSM_BAND_UNKNOWN;
-
-    /* If the modem has already been removed, return without
-     * scheduling callback */
-    if (mm_callback_info_check_modem_removed (info))
-        return;
-
-    if (error)
-        info->error = g_error_copy (error);
-    else if (parse_ipbm (response->str, &mm_band))
-        mm_callback_info_set_result (info, GUINT_TO_POINTER (mm_band), NULL);
-
-    mm_callback_info_schedule (info);
+    mm_modem_icera_set_band (MM_MODEM_ICERA (modem), band, callback, user_data);
 }
 
 static void
@@ -258,64 +113,15 @@ get_band (MMModemGsmNetwork *modem,
           MMModemUIntFn callback,
           gpointer user_data)
 {
-    MMAtSerialPort *port;
-    MMCallbackInfo *info;
-
-    info = mm_callback_info_uint_new (MM_MODEM (modem), callback, user_data);
-
-    /* Otherwise ask the modem */
-    port = mm_generic_gsm_get_best_at_port (MM_GENERIC_GSM (modem), &info->error);
-    if (!port) {
-        mm_callback_info_schedule (info);
-        return;
-    }
-
-    mm_at_serial_port_queue_command (port, "AT%IPBM?", 3, get_band_done, info);
+    mm_modem_icera_get_current_bands (MM_MODEM_ICERA (modem), callback, user_data);
 }
 
 static void
-send_samsung_pinnum_done (MMAtSerialPort *port,
-                          GString *response,
-                          GError *error,
-                          gpointer user_data)
+get_supported_bands (MMGenericGsm *gsm,
+                     MMModemUIntFn callback,
+                     gpointer user_data)
 {
-    MMCallbackInfo *info = (MMCallbackInfo *) user_data;
-    int matched;
-    GArray *retry_counts;
-    PinRetryCount ur[4] = {
-        {"sim-pin", 0}, {"sim-puk", 0}, {"sim-pin2", 0}, {"sim-puk2", 0}
-    };
-
-    /* If the modem has already been removed, return without
-     * scheduling callback */
-    if (mm_callback_info_check_modem_removed (info))
-        return;
-
-    if (error) {
-        info->error = g_error_copy (error);
-        goto done;
-    }
-
-    matched = sscanf (response->str, "%%PINNUM: %d, %d, %d, %d",
-                      &ur[0].count, &ur[1].count, &ur[2].count, &ur[3].count);
-    if (matched == 4) {
-        if (ur[0].count > 998) {
-            info->error = g_error_new (MM_MODEM_ERROR, MM_MODEM_ERROR_GENERAL,
-                                       "Invalid PIN attempts left %d", ur[0].count);
-            ur[0].count = 0;
-        }
-
-        retry_counts = g_array_sized_new (FALSE, TRUE, sizeof (PinRetryCount), 4);
-        g_array_append_vals (retry_counts, &ur, 4);
-        mm_callback_info_set_result (info, retry_counts, NULL);
-    } else {
-        info->error = g_error_new_literal (MM_MODEM_ERROR, MM_MODEM_ERROR_GENERAL,
-                                           "Could not parse PIN retries results");
-    }
-
-done:
-    mm_serial_port_close (MM_SERIAL_PORT (port));
-    mm_callback_info_schedule (info);
+    mm_modem_icera_get_supported_bands (MM_MODEM_ICERA (gsm), callback, user_data);
 }
 
 static void
@@ -340,31 +146,7 @@ get_unlock_retries (MMModemGsmCard *modem,
                     MMModemArrayFn callback,
                     gpointer user_data)
 {
-    MMAtSerialPort *port;
-    MMCallbackInfo *info = mm_callback_info_array_new (MM_MODEM (modem), callback, user_data);
-
-    mm_dbg ("get_unlock_retries");
-
-    /* Ensure we have a usable port to use for the command */
-    port = mm_generic_gsm_get_best_at_port (MM_GENERIC_GSM (modem), &info->error);
-    if (!port) {
-        mm_callback_info_schedule (info);
-        return;
-    }
-
-    /* Modem may not be enabled yet, which sometimes can't be done until
-     * the device has been unlocked.  In this case we have to open the port
-     * ourselves.
-     */
-    if (!mm_serial_port_open (MM_SERIAL_PORT (port), &info->error)) {
-        mm_callback_info_schedule (info);
-        return;
-    }
-
-    /* if the modem have not yet been enabled we need to make sure echoing is turned off */
-    mm_at_serial_port_queue_command (port, "E0", 3, NULL, NULL);
-    mm_at_serial_port_queue_command (port, "%PINNUM?", 3, send_samsung_pinnum_done, info);
-
+    mm_modem_icera_get_unlock_retries (MM_MODEM_ICERA (modem), callback, user_data);
 }
 
 static void
@@ -719,4 +501,5 @@ mm_modem_samsung_gsm_class_init (MMModemSamsungGsmClass *klass)
     gsm_class->set_allowed_mode = set_allowed_mode;
     gsm_class->get_allowed_mode = get_allowed_mode;
     gsm_class->get_access_technology = get_access_technology;
+    gsm_class->get_supported_bands = get_supported_bands;
 }
