@@ -50,6 +50,7 @@ typedef struct {
     gboolean has_net;
     gboolean is_icera;
     MMModemIceraPrivate *icera;
+    gboolean icera_dhcp;
 } MMModemZtePrivate;
 
 MMModem *
@@ -57,7 +58,8 @@ mm_modem_zte_new (const char *device,
                   const char *driver,
                   const char *plugin,
                   guint32 vendor,
-                  guint32 product)
+                  guint32 product,
+                  gboolean icera_dhcp)
 {
     MMModem *modem;
 
@@ -72,8 +74,10 @@ mm_modem_zte_new (const char *device,
                                     MM_MODEM_HW_VID, vendor,
                                     MM_MODEM_HW_PID, product,
                                     NULL));
-    if (modem)
+    if (modem) {
         MM_MODEM_ZTE_GET_PRIVATE (modem)->icera = mm_modem_icera_init_private ();
+        MM_MODEM_ZTE_GET_PRIVATE (modem)->icera_dhcp = icera_dhcp;
+    }
 
     return modem;
 }
@@ -118,13 +122,8 @@ get_allowed_mode_done (MMAtSerialPort *port,
         goto done;
     }
 
-    r = g_regex_new ("+ZSNT:\\s*(\\d),(\\d),(\\d)", G_REGEX_UNGREEDY, 0, NULL);
-    if (!r) {
-        info->error = g_error_new_literal (MM_MODEM_ERROR,
-                                           MM_MODEM_ERROR_GENERAL,
-                                           "Failed to parse the allowed mode response");
-        goto done;
-    }
+    r = g_regex_new ("\\+ZSNT:\\s*(\\d),(\\d),(\\d)", G_REGEX_UNGREEDY, 0, NULL);
+    g_assert (r != NULL);
 
     if (g_regex_match_full (r, response->str, response->len, 0, 0, &match_info, &info->error)) {
         MMModemGsmAllowedMode mode = MM_MODEM_GSM_ALLOWED_MODE_ANY;
@@ -430,22 +429,20 @@ icera_check_cb (MMModem *modem,
                 GError *error,
                 gpointer user_data)
 {
-    if (!error) {
+    if (!error && result) {
         MMModemZte *self = MM_MODEM_ZTE (user_data);
         MMModemZtePrivate *priv = MM_MODEM_ZTE_GET_PRIVATE (self);
 
-        if (result) {
-            priv->is_icera = TRUE;
+        priv->is_icera = TRUE;
 
-            /* Some devices with Icera chipsets don't have pseudo-ethernet
-             * ports and thus should use PPP.  Some devices use static IP,
-             * while others use DHCP on the net port.  What fun.
-             */
-            if (priv->has_net) {
-                g_object_set (G_OBJECT (modem),
-                              MM_MODEM_IP_METHOD, MM_MODEM_IP_METHOD_STATIC,
-                              NULL);
-            }
+        /* Some devices with Icera chipsets don't have pseudo-ethernet
+         * ports and thus should use PPP.  Some devices use static IP,
+         * while others use DHCP on the net port.  What fun.
+         */
+        if (priv->has_net) {
+            g_object_set (G_OBJECT (modem), MM_MODEM_IP_METHOD,
+                          priv->icera_dhcp ? MM_MODEM_IP_METHOD_DHCP : MM_MODEM_IP_METHOD_STATIC,
+                          NULL);
         }
     }
 }
@@ -670,6 +667,32 @@ port_grabbed (MMGenericGsm *gsm,
         MM_MODEM_ZTE_GET_PRIVATE (gsm)->has_net = TRUE;
 }
 
+static void
+set_property (GObject *object,
+              guint prop_id,
+              const GValue *value,
+              GParamSpec *pspec)
+{
+    /* Do nothing... see set_property() in parent, which also does nothing */
+}
+
+static void
+get_property (GObject *object,
+              guint prop_id,
+              GValue *value,
+              GParamSpec *pspec)
+{
+    switch (prop_id) {
+    case MM_GENERIC_GSM_PROP_POWER_DOWN_CMD:
+        /* Use AT+CFUN=4 for power down. It will stop the RF (IMSI detach), and
+         * keeps access to the SIM */
+        g_value_set_string (value, "+CFUN=4");
+        break;
+    default:
+        break;
+    }
+}
+
 /*****************************************************************************/
 
 static MMModemIceraPrivate *
@@ -751,6 +774,13 @@ mm_modem_zte_class_init (MMModemZteClass *klass)
     mm_modem_zte_parent_class = g_type_class_peek_parent (klass);
     g_type_class_add_private (object_class, sizeof (MMModemZtePrivate));
 
+    object_class->get_property = get_property;
+    object_class->set_property = set_property;
+
+    g_object_class_override_property (object_class,
+                                      MM_GENERIC_GSM_PROP_POWER_DOWN_CMD,
+                                      MM_GENERIC_GSM_POWER_DOWN_CMD);
+
     object_class->dispose = dispose;
     gsm_class->port_grabbed = port_grabbed;
     gsm_class->do_enable = do_enable;
@@ -759,4 +789,3 @@ mm_modem_zte_class_init (MMModemZteClass *klass)
     gsm_class->get_allowed_mode = get_allowed_mode;
     gsm_class->get_access_technology = get_access_technology;
 }
-
