@@ -45,11 +45,21 @@ enum {
     PROP_SEND_DELAY,
     PROP_FD,
     PROP_SPEW_CONTROL,
+    PROP_RTS_CTS,
     PROP_FLASH_OK,
 
     LAST_PROP
 };
 
+enum {
+    BUFFER_FULL,
+    TIMED_OUT,
+
+    LAST_SIGNAL
+};
+
+static guint signals[LAST_SIGNAL] = { 0 };
+
 #define SERIAL_BUF_SIZE 2048
 
 #define MM_SERIAL_PORT_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), MM_TYPE_SERIAL_PORT, MMSerialPortPrivate))
@@ -70,12 +80,15 @@ typedef struct {
     guint stopbits;
     guint64 send_delay;
     gboolean spew_control;
+    gboolean rts_cts;
     gboolean flash_ok;
 
     guint queue_id;
     guint watch_id;
     guint timeout_id;
 
+    guint n_consecutive_timeouts;
+
     guint flash_id;
     guint connected_id;
 } MMSerialPortPrivate;
@@ -313,7 +326,7 @@ static gboolean
 real_config_fd (MMSerialPort *self, int fd, GError **error)
 {
     MMSerialPortPrivate *priv = MM_SERIAL_PORT_GET_PRIVATE (self);
-    struct termios stbuf;
+    struct termios stbuf, other;
     int speed;
     int bits;
     int parity;
@@ -332,7 +345,7 @@ real_config_fd (MMSerialPort *self, int fd, GError **error)
                    errno);
     }
 
-    stbuf.c_iflag &= ~(IGNCR | ICRNL | IUCLC | INPCK | IXON | IXANY | IGNPAR );
+    stbuf.c_iflag &= ~(IGNCR | ICRNL | IUCLC | INPCK | IXON | IXANY );
     stbuf.c_oflag &= ~(OPOST | OLCUC | OCRNL | ONLCR | ONLRET);
     stbuf.c_lflag &= ~(ICANON | XCASE | ECHO | ECHOE | ECHONL);
     stbuf.c_lflag &= ~(ECHO | ECHOE);
@@ -340,14 +353,34 @@ real_config_fd (MMSerialPort *self, int fd, GError **error)
     stbuf.c_cc[VTIME] = 0;
     stbuf.c_cc[VEOF] = 1;
 
-    /* Use software handshaking */
-    stbuf.c_iflag |= (IXON | IXOFF | IXANY);
+    /* Use software handshaking and ignore parity/framing errors */
+    stbuf.c_iflag |= (IXON | IXOFF | IXANY | IGNPAR);
 
     /* Set up port speed and serial attributes; also ignore modem control
      * lines since most drivers don't implement RTS/CTS anyway.
      */
     stbuf.c_cflag &= ~(CBAUD | CSIZE | CSTOPB | PARENB | CRTSCTS);
-    stbuf.c_cflag |= (speed | bits | CREAD | 0 | parity | stopbits | CLOCAL);
+    stbuf.c_cflag |= (bits | CREAD | 0 | parity | stopbits | CLOCAL);
+
+    errno = 0;
+    if (cfsetispeed (&stbuf, speed) != 0) {
+        g_set_error (error,
+                     MM_MODEM_ERROR,
+                     MM_MODEM_ERROR_GENERAL,
+                     "%s: failed to set serial port input speed; errno %d",
+                     __func__, errno);
+        return FALSE;
+    }
+
+    errno = 0;
+    if (cfsetospeed (&stbuf, speed) != 0) {
+        g_set_error (error,
+                     MM_MODEM_ERROR,
+                     MM_MODEM_ERROR_GENERAL,
+                     "%s: failed to set serial port output speed; errno %d",
+                     __func__, errno);
+        return FALSE;
+    }
 
     if (tcsetattr (fd, TCSANOW, &stbuf) < 0) {
         g_set_error (error,
@@ -358,6 +391,22 @@ real_config_fd (MMSerialPort *self, int fd, GError **error)
         return FALSE;
     }
 
+    /* tcsetattr() returns 0 if any of the requested attributes could be set,
+     * so we should double-check that all were set and log a warning if not.
+     */
+    memset (&other, 0, sizeof (struct termios));
+    errno = 0;
+    if (tcgetattr (fd, &other) != 0) {
+        mm_warn ("(%s): tcgetattr() error: %d",
+                 mm_port_get_device (MM_PORT (self)),
+                 errno);
+    }
+
+    if (memcmp (&stbuf, &other, sizeof (other)) != 0) {
+        mm_warn ("(%s): port attributes not fully set",
+                   mm_port_get_device (MM_PORT (self)));
+    }
+
     return TRUE;
 }
 
@@ -417,6 +466,10 @@ mm_serial_port_process_command (MMSerialPort *self,
         if (errno == EAGAIN || status == 0) {
             info->eagain_count--;
             if (info->eagain_count <= 0) {
+                /* If we reach the limit of EAGAIN errors, treat as a timeout error. */
+                priv->n_consecutive_timeouts++;
+                g_signal_emit (self, signals[TIMED_OUT], 0, priv->n_consecutive_timeouts);
+
                 g_set_error (error, MM_SERIAL_ERROR, MM_SERIAL_ERROR_SEND_FAILED,
                              "Sending command failed: '%s'", strerror (errno));
                 return FALSE;
@@ -544,14 +597,22 @@ mm_serial_port_timed_out (gpointer data)
 
     priv->timeout_id = 0;
 
+    /* Update number of consecutive timeouts found */
+    priv->n_consecutive_timeouts++;
+
     error = g_error_new_literal (MM_SERIAL_ERROR,
                                  MM_SERIAL_ERROR_RESPONSE_TIMEOUT,
                                  "Serial command timed out");
+
     /* FIXME: This is not completely correct - if the response finally arrives and there's
        some other command waiting for response right now, the other command will
        get the output of the timed out command. Not sure what to do here. */
     mm_serial_port_got_response (self, error);
 
+    /* Emit a timed out signal, used by upper layers to identify a disconnected
+     * serial port */
+    g_signal_emit (self, signals[TIMED_OUT], 0, priv->n_consecutive_timeouts);
+
     return FALSE;
 }
 
@@ -678,12 +739,15 @@ data_available (GIOChannel *source,
         /* Make sure the response doesn't grow too long */
         if ((priv->response->len > SERIAL_BUF_SIZE) && priv->spew_control) {
             /* Notify listeners and then trim the buffer */
-            g_signal_emit_by_name (self, "buffer-full", priv->response);
+            g_signal_emit (self, signals[BUFFER_FULL], 0, priv->response);
             g_byte_array_remove_range (priv->response, 0, (SERIAL_BUF_SIZE / 2));
         }
 
-        if (parse_response (self, priv->response, &err))
+        if (parse_response (self, priv->response, &err)) {
+            /* Reset number of consecutive timeouts only here */
+            priv->n_consecutive_timeouts = 0;
             mm_serial_port_got_response (self, err);
+        }
     } while (bytes_read == SERIAL_BUF_SIZE || status == G_IO_STATUS_AGAIN);
 
     return TRUE;
@@ -902,6 +966,7 @@ mm_serial_port_close (MMSerialPort *self)
                                          "Serial port is now closed");
             response = g_byte_array_sized_new (1);
             g_byte_array_append (response, (const guint8 *) "\0", 1);
+
             MM_SERIAL_PORT_GET_CLASS (self)->handle_response (self,
                                                               response,
                                                               error,
@@ -1041,6 +1106,7 @@ get_speed (MMSerialPort *self, speed_t *speed, GError **error)
 static gboolean
 set_speed (MMSerialPort *self, speed_t speed, GError **error)
 {
+    MMSerialPortPrivate *priv = MM_SERIAL_PORT_GET_PRIVATE (self);
     struct termios options;
     int fd, count = 4;
     gboolean success = FALSE;
@@ -1061,6 +1127,10 @@ set_speed (MMSerialPort *self, speed_t speed, GError **error)
     cfsetospeed (&options, speed);
     options.c_cflag |= (CLOCAL | CREAD);
 
+    /* Configure flow control as well here */
+    if (priv->rts_cts)
+        options.c_cflag |= (CRTSCTS);
+
     while (count-- > 0) {
         if (tcsetattr (fd, TCSANOW, &options) == 0) {
             success = TRUE;
@@ -1315,6 +1385,9 @@ set_property (GObject *object, guint prop_id,
     case PROP_SPEW_CONTROL:
         priv->spew_control = g_value_get_boolean (value);
         break;
+    case PROP_RTS_CTS:
+        priv->rts_cts = g_value_get_boolean (value);
+        break;
     case PROP_FLASH_OK:
         priv->flash_ok = g_value_get_boolean (value);
         break;
@@ -1356,6 +1429,9 @@ get_property (GObject *object, guint prop_id,
     case PROP_SPEW_CONTROL:
         g_value_set_boolean (value, priv->spew_control);
         break;
+    case PROP_RTS_CTS:
+        g_value_set_boolean (value, priv->rts_cts);
+        break;
     case PROP_FLASH_OK:
         g_value_set_boolean (value, priv->flash_ok);
         break;
@@ -1469,6 +1545,14 @@ mm_serial_port_class_init (MMSerialPortClass *klass)
                                FALSE,
                                G_PARAM_READWRITE));
 
+    g_object_class_install_property
+        (object_class, PROP_RTS_CTS,
+         g_param_spec_boolean (MM_SERIAL_PORT_RTS_CTS,
+                               "RTSCTS",
+                               "Enable RTS/CTS flow control",
+                               FALSE,
+                               G_PARAM_READWRITE));
+
     g_object_class_install_property
         (object_class, PROP_FLASH_OK,
          g_param_spec_boolean (MM_SERIAL_PORT_FLASH_OK,
@@ -1479,12 +1563,22 @@ mm_serial_port_class_init (MMSerialPortClass *klass)
                                G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
 
     /* Signals */
-    g_signal_new ("buffer-full",
+    signals[BUFFER_FULL] =
+        g_signal_new ("buffer-full",
                   G_OBJECT_CLASS_TYPE (object_class),
                   G_SIGNAL_RUN_FIRST,
                   G_STRUCT_OFFSET (MMSerialPortClass, buffer_full),
                   NULL, NULL,
                   g_cclosure_marshal_VOID__POINTER,
                   G_TYPE_NONE, 1, G_TYPE_POINTER);
+
+    signals[TIMED_OUT] =
+        g_signal_new ("timed-out",
+                      G_OBJECT_CLASS_TYPE (object_class),
+                      G_SIGNAL_RUN_FIRST,
+                      G_STRUCT_OFFSET (MMSerialPortClass, timed_out),
+                      NULL, NULL,
+                      g_cclosure_marshal_VOID__UINT,
+					  G_TYPE_NONE, 1, G_TYPE_UINT);
 }
 
